unit xMinR1SearchR1;
{ Feel free to use this unit as you want.
  First developped by Bernard Gisin

  This unit implements different minimisation algorithms
  of a function : R --> R of one variable

->procedure SearchMinGoldenSection;
  Minimisation using the Golden Section Search in one dimension
  See "Numerical recipes" Second edition Chap. 10.1,  page 397

->procedure SearchMinBrent;
  Minimisation using the Parabolic interpolation and Brent's method in one dimension
  See "Numerical recipes" Second edition Chap. 10.2,  page 402
  The Idea of the algorithme is to go quickly to the minimum using the minimum
  of a parabolic fit, and when it seems to be reached, because the parabole
  minimum is near the actual minimum, then evaluate the function left
  and right of the minimum to be sure that it has been bracketed.

  See tEx066.pp for examples and tests.

  A Little theory :
  """""""""""""""""
  Suppose that  F(x) = (x + 10^3)^2 - 2*10^3*x - 10^6
  Mathematically it is equivalent to  F(x) = x^2,  but numerically, it is
  different. We have :
  F(x) = 10^6(1 + x^2/10^6) - 10^6
  any number  x  such that  abs(x^2/10^6) < gglvFloatEpsilon will give  F(x) = 0
  where  gglvFloatEpsilon  is defined in  wGlobals2.pp and
  is the smallest positive number such that  (1.0 + gglvFloatEpsilon) <> 1.0

  Let see it more generally
  F(x) = y0 + a*(x-x0)^2
  F(x+e) = y0 + a*(x-x0)^2 * (1 + 2*e/(x-x0) + e^2/(x-x0)^2)
         = y0 + a*(x-x0)^2 * (1 + 2*e/(x-x0))
         = F(x) + 2*e*(x-x0)
         if abs(e) < (x-x0) * sqrt(gglvFloatEpsilon)
  # If  x=x0  then it is equal to  F(x).
  This show that the minimum cannot be found
  at a precision better than : (x-x0) * sqrt(gglvFloatEpsilon)
  # If  x <> x0  then it is still different from  F(x)  but
  F(x), F(x+e),  F(x+2*e) will lie on a line and parabolic interpolation
  will give no minimum.

  Conclusion,  Evaluating  F  at two numbers which difference are smaller
  than  (x-x0) * sqrt(gglvFloatEpsilon)  is useless. It will not give a better
  value for the minimum, and it will not give a good information to
  compute the parabolic interpolation.

  There is a second restriction :
  F(x) = y0 * (1 + a*(x-x0)^2/y0)
  thus  abs(a*(x-x0)^2/y0) > gglvFloatEpsilon  to have a value different from  y0
  i.e.  x-x0 > sqrt(y0 * gglvFloatEpsilon / a)

  Conclusion,  evaluating  F  at two numbers which difference are smaller than
  sqrt(y0 * gglvFloatEpsilon / a)  is useless.

  The example  F(x) = 10^6(1 + x^2/10^6) - 10^6  show that both conditions
  above are necessary, but not enough. In this example,  x0=y0=0.
  This is why I will take  at least  1.0  for  x0  and for  y0
  and a security factor.
  ----------------------------------------------------------------}

INTERFACE
//"""""""

uses wGlobals2;

type
  PMinR1SearchR1 = ^TMinR1SearchR1;
  TMinR1SearchR1 = object
  //==========================
  // Must be set before calling some minimisation procedure. See their comments
  vXL,  // left  of the searched interval
  vXR,  // right of the searched interval
  vXM,  // in the middle somewhere
  vFL,  // functions evaluations at the corresponding  x  values
  vFM,
  vFR
         : Float;

  Func : TypeFuncR1R1Data;
  // Function to minimise. Must be set before calling a minimisation algorithme

  // Optional Inputs :
  //""""""""""""""""""
  vTolAbs : Float;
  // When the difference  vXR-vXL <= vTolAbs, stop the search
  // Do not set vTolAbs < sqrt(machine precision), it is useless,
  // i.e. vTolAbs >= 1e-8 is resonable, bellow it is not.

  nNbFuncEvalMax : int32;
  // Maximum number of functions evaluations accepted

  pDataFunc : pointer;
  // Pointer passed to the function when it is evaluated.

  wOptions : word32; // Can be "or" combinaison of the following
  // =1 => vFL=Func(vXL) has been computed, do not recompute it
  // =2 => vFR=Func(vXR) has been computed, do not recompute it
  // =4 => vFM=Func(vXM) has been computed, do not recompute it

  // Results :
  //""""""""""
  nExitCode : integer;
  // = 0 when the minimum has been found, i.e. the vTolEnd condition has been reached
  // = 1 when the maximum number of iterations has been reached.
  // = 2 when no minimum has been found in the starting interval [vXL, vXR]
  // = 3 when the starting point don't backet a minimum

  nNbFuncEval : int32;
  // number of function evaluations

  vXSol : Float;
  // Solution found where the function has been evaluated

  vFSol : Float;
  // evaluation of the function at the minimum found : vXSol

  vXFMin : Float;
  // Value at which the minium has been found.
  // The function has not been evaluated there, but it is the minimum
  // of the parabola passing through  (vXL, vFL), (vXM, vFM), (vXR, vFR)

  constructor Init;
  destructor Done; virtual;

  procedure GetParabole(vX1,vX2,vX3,vF1,vF2,vF3:Float; var vAV,vXMinV,vYMinV:Float);
  // Compute the parabola : y = vAV * (x - vXMinV)^2 + vYMin
  // passing through the points (vXj, vFj) j = 1,2,3
  // If vAV < 0 then vYMin is the maximum of the parabola.
  // If vAV = 0 then the points are on a line, and vXMinV=(vX1+vX2+vX3)/3 &  vYMinV=(vY1+vY2+vY3)/3

  procedure SearchMinBracket(vStepSearch:Float);
  // Try to find three values : vXL <  vXM <  vXR  such that
  // Max(Func(vXL), Func(vXR)) <= Func(vXM)  (Left, Middle, Right)
  // Inputs :
  //"""""""""
  // vXL do not search  X < vXL
  // vXR do not search  X > vXR
  // vXM = starting search value
  // vStepSearch = step to search the minimum. Can be negative
  // Func
  // Optional inputs :
  //""""""""""""""""""
  // nNbFuncEvalMax
  // wOptions
  // vFM
  // Outputs :
  //"""""""""
  // nExitCode=0 => found OK
  // nExitCode=2 => not found
  // nNbFuncEval = number of functions evaluation done
  // vXL < vXM < vXR such that Func(vXL)=vFL >= Func(vXM)=vFM <= Func(vXR)=vFR

  procedure SearchMinGoldenSection;
  // It is not a very good algorithme, but if a minimum has been bracketed,
  // then with certitude, it will converge to a minimum.
  // Inputs :
  //"""""""""
  // vXL, vXM, vXR
  // Func
  // Optional inputs :
  //""""""""""""""""""
  // nNbFuncEvalMax
  // vTolAbs
  // wOptions
  // vFL,vFM,vFR
  // Output :
  //"""""""""
  // nExitCode=0 => found OK
  // nExitCode=1 => maximum number of iteration reached.
  // nNbFuncEval = number of functions evaluation done
  // vXSol, vFSol
  // It is possible that vXSol = one of the starting extremities,
  // signifying that the minimum was in a border
  // vXFMin = Minimum. The function has not been evaluated there

  procedure SearchMinBrent;
  // A good algorithme.
  // Inputs :
  //"""""""""
  // vXL, vXM, vXR
  // Func
  // Optional inputs :
  //""""""""""""""""""
  // nNbFuncEvalMax
  // vTolAbs
  // wOptions
  // vFL,vFM,vFR
  // Output :
  //"""""""""
  // nExitCode=0 => found OK
  // nExitCode=1 => maximum number of iteration reached.
  // nNbFuncEval = number of functions evaluation done
  // vXSol = point at which the minimum function evaluation has been done
  // vFSol = Func(vXSol)
  // It is possible that vXSol = one of the starting extremities,
  // signifying that the minimum was in a border
  // vXFMin = Minimum. The function has not been evaluated there

  end; // TMinR1SearchR1 = object

IMPLEMENTATION
//""""""""""""

Uses {wWindowText0,} wDebug0;

var nNotUsed : integer; // only to avoid an error message while compliling in DevPascal 1.9

//****************************************************************
// TMinR1SearchR1 = object
//****************************************************************

constructor TMinR1SearchR1.Init;
//============================
begin
// This must still be set
// Func = @FunctionToMin
// vXL, vXR,  vXM sometimes

// This can be changed
vTolAbs   :=1e-3; // This is usually enough, the vXFMin will usually be much better
nNbFuncEvalMax:=5000;
pDataFunc :=NIL;

wOptions  :=0;
end;  //  TMinR1SearchR1.Init

destructor TMinR1SearchR1.Done;
//===========================
begin
end;  //  TMinR1SearchR1.Done

procedure TMinR1SearchR1.GetParabole(vX1,vX2,vX3,vF1,vF2,vF3:Float; var vAV,vXMinV,vYMinV:Float);
//=============================================================================================
// Compute the parabola : y = vAV * (x - vXMinV)^2 + vYMin
// passing through the points (vXj, vFj) j = 1,2,3
var
  vB12,vB23 : Float;  // divided differences
begin
if (vX1 = vX2) or (vX2 = vX3) or (vX1 = vX3) then begin
   // only two points, no parabola exist.
   vAV:=0;
   vXMinV:=(vX1+vX2+vX3)/3;
   vYMinV:=(vF1+vF2+vF3)/3;
   exit;
   end;

// Compute the divided differences, to get the parabola
vB12:=(vF2 - vF1) / (vX2-vX1);
vB23:=(vF3 - vF2) / (vX3-vX2);
vAV:=(vB23 - vB12) / (vX3 - vX1);

if (abs(vAV) <= 1e-8) then begin
   // The points are on a line
   vAV:=0;
   vXMinV:=(vX1+vX2+vX3)/3;
   vYMinV:=(vF1+vF2+vF3)/3;
   exit;
   end;

vXMinV:=0.5*(vX1 + vX2 - vB12 / vAV); // = 0.5*(vX2 + vX3 - vB23 / vAV);
vYMinV:=(vAV * (vXMinV - vX2) + vB12) * (vXMinV - vX1) + vF1;
end;  //  TMinR1SearchR1.GetParabole

procedure TMinR1SearchR1.SearchMinBracket(vStepSearch:Float);
//=========================================================
// Try to find three values : vXL <  vXM <  vXR  such that
// Max(Func(vXL), Func(vXR)) <= Func(vXM)  (Left, Middle, Right)
// Inputs :
// vXL do not search  X < vXL
// vXR do not search  X > vXR
// vXM = starting search value
// vStepSearch = step to search the minimum. Can be negative
// Outputs :
// vXL < vXM < vXR such that Func(vXL)=vFL >= Func(vXM)=vFM <= Func(vXR)=vFR
var
  vXHi,       // a point such that Func(vXHi) > Func(vXM) = vFM
  vXnext,     // next trial point not betewwn vXM and vXHi,
              // to get Func(vXnext) > Func(vFM) and thus a bracketing
  vFHi,       // =Func(vFHi)
  vFnext,     // =Func(vFnext)
  vAParab,    // For parabolic interpolation
  vYMinParab,
  vXMinParab,
  vFMinParab, // evaluation of Func at the min of the parabola
  vXnextTry,  // next trial point to evaluate the function
  vTemp
              : Float;
begin
nExitCode  :=2;
nNbFuncEval:=0;

// Evaluating the function at the starting points
if ((wOptions and 4) = 0) then begin vFM :=Func(pDataFunc, vXM); inc(nNbFuncEval); end;

vXHi:=vXM + vStepSearch;
vFHi:=Func(pDataFunc, vXHi);  inc(nNbFuncEval);

if (vFHi < vFM) then begin
   // Switch vFM and vFHi such that  vFHi >= vFM
   vTemp :=vXHi;
   vXHi  :=vXM;
   vXM   :=vTemp;
   vTemp :=vFHi;
   vFHi  :=vFM;
   vFM   :=vTemp;
   end
else vStepSearch:=-vStepSearch; // The search direction changes

// Here vFHi >= vFM and  vXM + vStepSearch  is bigger or smaller than both vXM and vXHi

vXnext :=vXM + vStepSearch;
repeat
   vFnext :=Func(pDataFunc, vXnext);  inc(nNbFuncEval);

//DBW(16, FTS(vXHi) + FTS(vXM) + FTS(vXnext) + FTS(vFHi) + FTS(vFM) + FTS(vFnext) );
   if (vFnext > vFM) then begin
      // Bracket found
      // Arrange the values to return the result in vXL < vXM < vXR
      if (vXHi > vXnext) then begin
         vXL:=vXnext;  vFL:=vFnext;
         vXR:=vXHi;    vFR:=vFHi;
         end
      else begin // vXHi <= vXnext
         vXR:=vXnext;  vFR:=vFnext;
         vXL:=vXHi;    vFL:=vFHi;
         end;

      // vXM, vFM  are already correct.
      nExitCode :=0;
      exit;
      end;

   // Next trial function evaluation
   vXnextTry:=vXnext + vStepSearch;

   // Try to find a minimum by parabolic interpolation
   GetParabole(vXHi,vXM,vXnext,vFHi,vFM,vFnext, vAParab,vXMinParab,vYMinParab);
   if (vAParab > 0) then begin // If not, the we got a maximum, not a minimum
      if ((vXM + vTolAbs < vXMinParab) and (vXMinParab < vXnext - vTolAbs)) or
         ((vXnext + vTolAbs < vXMinParab) and (vXMinParab < vXM - vTolAbs)) then begin
         // Here we expect the minimum to be tbetween vXM and vXnext
         vFMinParab :=Func(pDataFunc, vXMinParab);  inc(nNbFuncEval);

//DBW(16, 'Expecting a min at parabola min' + FTS(vXMinParab) + FTS(vFMinParab));

         if (vFMinParab < vFnext) then begin
            // Finished, we bracketed the minimum
            if (vXHi > vXnext) then begin
               vXL:=vXnext;  vFL:=vFnext;
               vXR:=vXM;     vFR:=vFM;
               end
            else begin // vXHi <= vXnext
               vXR:=vXnext;  vFR:=vFnext;
               vXL:=vXM;     vFL:=vFM;
               end;

            vXM:=vXMinParab; vFM:=vFMinParab;
            nExitCode :=0;
            exit;
            end
         else if (vFMinParab > vFM) then begin
            // Not expected, quite strange, but we got a minimum
            if (vXHi > vXnext) then begin
               vXL:=vXMinParab; vFL:=vFMinParab;
               vXR:=vXHi;       vFR:=vFHi;
               end
            else begin // vXHi <= vXnext
               vXR:=vXMinParab;  vFR:=vFMinParab;
               vXL:=vXHi;        vFL:=vFHi;
               end;

            nExitCode :=0;
            exit;
            end
         else begin
            // No minimum found, we can just move in the search interval
            // This will be changed just bellow, under "Try to go further"
            vXM:=vXMinParab;  vFM:=vFMinParab;
            end;
         end // if ((vXM + vTolAbs < vXMinParab) and (vXMinParab < vXnext - vTolAbs)) or
{ Give very often bad results
      else begin
         // the minimum of the parabola is further than vXnext,
         // But maybe it gives a good point to try
         if ((vXnext + vTolAbs < vXMinParab) and
             (vXMinParab < vXnext + 0.5*vStepSearch)) or
            ((vXnext + 0.5*vStepSearch < vXMinParab) and
             (vXMinParab < vXnext - vTolAbs)) then begin
            // Try to advance less than the usual vStepSearch
            // Expecting that after two more function evaluation, we got the min.
            vXnextTry:=2*vXMinParab - vXNext;
//DBW(16, 'Changing the next trial point' + FTS(vXnextTry) + FTS(vXMinParab));
            end;
         end;
}
      end; // if (vAParab > 0) then begin

   // Try to go further
   vXHi   :=vXM;      vFHi:=vFM;
   vXM    :=vXnext;   vFM :=vFnext;
   vXNext :=vXnextTry;

   if (vXNext < vXL) and (vXM > vXL) then vXnext:=vXL; // Last try
   if (vXNext > vXR) and (vXM < vXR) then vXnext:=vXR; // Last try
   until (vXnext < vXL) or (vXNext > vXR) or (nNbFuncEval >= nNbFuncEvalMax);

// No minimum has been found in the starting interval [vXL, vXR]
nExitCode :=2;

if (vXnext < vXL) then begin
   vXR:=vXHi;    vFR:=vFHi;
                 vFL:=vFnext;  // Not equal to Func(vXL) !!!
   end
else begin
   vXL:=vXHi;    vFL:=vFHi;
                 vFR:=vFnext;  // Not equal to Func(vXR) !!!
   end;
end;  // TMinR1SearchR1.SearchMinBracket

procedure TMinR1SearchR1.SearchMinGoldenSection;
//==============================================
// The initial conditions  vXL < vXM < vXR  such that
// Max(Func(vXL), Func(vXR)) <= Func(vXM)  (Left, Middle, Right)
// must be given.
// Will find a minimum of the function between vXL and vXR
var
  vXML, // Middle left
  vXMR, // Middle right
  vFML, // evaluation of the function at the corresponding  x
  vFMR,
  vAParab,
  vYMinParab  : Float;

begin
nNbFuncEval:=0;

if ((wOptions and 1) = 0) then begin vFL :=Func(pDataFunc, vXL);  inc(nNbFuncEval); end;
if ((wOptions and 2) = 0) then begin vFR :=Func(pDataFunc, vXR);  inc(nNbFuncEval); end;
if ((wOptions and 4) = 0) then begin vFM :=Func(pDataFunc, vXM);  inc(nNbFuncEval); end;

// We will always have  vXL < vXML < vXMR < vXR
// Use a fourth point in the largest interval
if (vXR - vXM) > (vXM - vXL) then begin
   // new point between  vXM and vXR
   vXML:=vXM;  vFML:=vFM;
   vXMR:=0.382*vXML + 0.618*vXR;
   vFMR:=Func(pDataFunc, vXMR);   inc(nNbFuncEval);
   end
else begin
   // new point between  vXL and vXM
   vXMR:=vXM;  vFMR:=vFM;
   vXML:=0.618*vXL + 0.382*vXMR;
   vFML:=Func(pDataFunc, vXML);   inc(nNbFuncEval);
   end;

repeat
   if (vFML > vFMR) then begin
      vXL :=vXML;  vFL :=vFML;
      vXML:=vXMR;  vFML:=vFMR;
      vXMR:=0.618*vXML + 0.382*vXR;
      vFMR:=Func(pDataFunc, vXMR);   inc(nNbFuncEval);
      end
   else begin
      vXR :=vXMR;  vFR :=vFMR;
      vXMR:=vXML;  vFMR:=vFML;
      vXML:=0.382*vXL + 0.618*vXMR;
      vFML:=Func(pDataFunc, vXML);   inc(nNbFuncEval);
      end;

//DBW(1, FTS(vXL) + FTS(vXML) + FTS(vXMR) + FTS(vXR));
//DBW(1, FTS((vXML-vXL)/(vXR-vXL)) + FTS((vXMR-vXL)/(vXR-vXL)));
   until (vXR - vXL <= vTOLAbs) or (nNbFuncEval >= nNbFuncEvalMax);

// Minimum value found
if (vFML < vFMR) then begin
   vXM   :=vXML;
   vFM   :=VFML;
   vXSol :=vXML;
   vFSol :=VFML;
   end
else begin
   vXM   :=vXMR;
   vFM   :=VFMR;
   vXSol :=vXMR;
   vFSol :=VFMR;
   end;

// Get a better minimum by parabolic interpoalation
GetParabole(vXL,vXM,vXR,vFL,vFM,vFR, vAParab,vXFMin,vYMinParab);
if (vAParab <= 0) or (vXFMin < vXL) or (vXFMin > vXR) then
   vXFMin:=vXM; // The parabola doesn't give a better min

if (nNbFuncEval >= nNbFuncEvalMax) then nExitCode :=1
else                                    nExitCode :=0;
end;  // TMinR1SearchR1.SearchMinGoldenSection;

procedure TMinR1SearchR1.SearchMinBrent;
//======================================
var
  vDX,         // variation of X to compute the next Functon evaluation
  vXStepSize,  // magnitude of last step size
  vXStepSizeLast, // Last value of  vXStepSize
  vFMin,       // Minimal value achieved so far
  vFMinSec,    // Second minimal value achieved so far
  vFMinSecLast,// Last value of vMinSec
  vFRecent,    // Most recent evaluation of Func
  vXMin,       // Point with the minimal function evaluation
  vXMinSec,    // Point with the second minimal function evaluation
  vXMinSecLast,// Last value of vXMinSec
  vXRecent,    // Most recent value of x at which the function has been evaluated
  vXMiddle,    // Middle of the interval  [vXL, vXR]
  vTemp,
  vNum1,
  vNum2,       // To compute the X of the minimum parabolic fit
  vNum,        // Numerator   to compute the parabolic fit
  vDen,        // Denominator to compute the parabolic fit
  vAParab,
  vYMinParab,
  vTolEps      // never evaluate the function at two values which difference is
               // less than vTolEps.
              : Float;
  nGood       // Number of last iterations which was done by parabolic fit
              // and gave a delta_X less than a fourth of the last delta_X
              : integer;
begin
nNbFuncEval:=0;
vDX:=1;  // Just to avoid a warning

// Assure that vXL < vXR  It is usually assumed to be so.
if (vXL > vXR) then begin
   vTemp:=vXL;
   vXL  :=vXR;
   vXR  :=vTemp;
   end;

// Starting points.
vXMinSecLast := vXM;
vXMinSec     := vXMinSecLast;
vXMin        := vXMinSecLast;
vXStepSize   := 0.0;  // No last step sizes

if ((wOptions and 4) = 0) then begin vFM :=Func(pDataFunc, vXM); inc(nNbFuncEval); end;
vFMinSecLast := vFM;
vFMinSec     := vFM;
vFMin        := vFM;

// If function evaluation has already be done at some points, use them
if ((wOptions and 3) = 3) then begin
   // vFL =Func(pDataFunc, vXL) and  vFR =Func(pDataFunc, vXR)  are known.
   // We know also that  vFM <= Min(vFL, vFR)
//   vXMinSecLast := vXM;   vFMinSecLast := vFMin;  // Already done
   if (vFL < vFR) then begin
      vXMinSecLast := vXR;   vFMinSecLast:= vFR;
      vXMinSec     := vXL;   vFMinSec    := vFL;
      end
   else begin
      vXMinSecLast := vXL;   vFMinSecLast := vFL;
      vXMinSec     := vXR;   vFMinSec     := vFR;
      end;

   vDX           := vXR - vXL;  // No restriction on the last step size
   vXStepSize    := vDX;
   vXStepSizeLast:= vXStepSize;
   end
else if ((wOptions and 1) = 1) then begin
   // vFL =Func(pDataFunc, vXL)  is known.
   vXMinSec   := vXL;   vFMinSec := vFL;
   vDX        := vXR - vXL;
   vXStepSize :=0;
   vFR :=0; // just to be initialised
   end
else if ((wOptions and 2) = 2) then begin
   // vFR =Func(pDataFunc, vXR)  is known.
   vXMinSec   := vXR;   vFMinSec := vFR;
   vDX        := vXR - vXL;
   vXStepSize :=0;
   vFL :=0; // just to be initialised
   end
else begin
   vFL :=0; // just to be initialised
   vFR :=0;
   end;

//DBW(16, FTS(vXL) + FTS(vXMin) + FTS(vXR) + FTS(vFMin));

// Main loop to bracket the minimum
nGood:=0;
repeat
//DBW(16, '-------------------------------');
   vXMiddle := 0.5*(vXL+vXR);
   if (vXR-vXL <= vTolAbs) then begin
      // Finished, we found the minimum
      vXM  := vXMin;
      vFM  := vFMin;
      vXSol:= vXMin;
      vFSol:= vFMin;
//DBW(16, 'End with :' + FTS(vXL) + FTS(vXM) + FTS(vXR));
//DBW(16, 'F=        ' + FTS(vFL) + FTS(vFM) + FTS(vFR));

      // Get a better minimum by parabolic interpoalation
      GetParabole(vXL,vXM,vXR,vFL,vFM,vFR, vAParab,vXFMin,vYMinParab);
      if (vAParab <= 0) or (vXFMin < vXL) or (vXFMin > vXR) then
         vXFMin:=vXM; // The parabola doesn't give a better min

      nExitCode :=0;
      exit;
      end;

   // No evaluations should be made at distance less than  vTolEps
   // 64.0 is a securities to be not too near from the limit
   vTolEps:=64*(abs(vXM) + 1.0) * sqrt(gglvFloatEpsilon);
   vTemp:=64*sqrt((abs(vFMin)+1.0) * gglvFloatEpsilon / (abs(vAParab)+1.0));
   if (vTolEps < vTemp) then vTolEps:=vTemp;

   if (abs(vXStepSize) > vTolEps) then begin
      // We can try the parabolic fit
      vNum1 := (vXMin-vXMinSec)*(vFMin-vFMinSecLast);
      vNum2 := (vXMin-vXMinSecLast)*(vFMin-vFMinSec);
      vNum  := (vXMin-vXMinSecLast)*vNum2-(vXMin-vXMinSec)*vNum1;
      vDen  := 2.0*(vNum2-vNum1);

      // We want the denominator to be positif
      if (vDen > 0.0) then vNum:= -vNum;
      vDen := abs(vDen);
      vXStepSizeLast := vXStepSize;
      vXStepSize := vDX;

      // vXStepSizeLast = maximum of last step and step before last
      if (abs(vXStepSizeLast) < abs(vDX)) then vXStepSizeLast:=vDX;
//DBW(16, 'vDX=' + FTS(vDX) + FTS(vXStepSizeLast));

      // Test if we use the parabole minimum
      // The minimum must satisfy :
      // 1) It must be in the interval [vXL..vXR] (two last conditions) if the IF
      // 2) It must move less than half the step before last step or
      //                         before the last step
      if (abs(vNum) >= abs(0.5*vDen*vXStepSizeLast)) or
         (vNum <= vDen*(vXL-vXMin)) or
         (vNum >= vDen*(vXR-vXMin)) then begin
         // Parabolic fit will not be used,
         // take the golden section in the larger of
         // the intervals [vXL,vXMin]; [vXMin,vXR]
         if (vXMin >= vXMiddle) then vXStepSize:= vXL-vXMin
         else                        vXStepSize:= vXR-vXMin;
         vDX := 0.381966*vXStepSize;
//DBW(16, 'No parab, Golden, vXRec=' + FTS(vXMin + vDX) + FTS(vXMin + vNum/(vDen+1.7e-9)));
         nGood:=0;
         end
      else begin
         // Use the minimum of the parabolic fit
         vDX := vNum/vDen;
         vXRecent := vXMin+vDX;

         // Do not evaluate too near from the interval limit
         // One limit can be vary near the minimum, but
         // the other must remain at a distance at least equal to 0.4*vTolAbs
         if (vXRecent - vXL < 0.4*vTolAbs) then vDX:= 0.4*vTolAbs;
         if (vXR - vXRecent < 0.4*vTolAbs) then vDX:=-0.4*vTolAbs;

         if (nGood >= 3) and (abs(vDX) < vTolAbs) and
            (((vXMin - vXL) < vTolAbs) or ((vXR - vXMin) < vTolAbs)) then begin
            // The next evaluation of func is useless, we are near enough
            // from the min. It is just done to bracket well the minimum.
            vXM  := vXMin;
            vFM  := vFMin;
            vXSol:= vXMin;
            vFSol:= vFMin;
            vXFMin:=vXM;

            // vXL  or  vXR  are not near the solution, but we don't care,
            // We are near enough from the solution
//DBW(16, 'End with :' + FTS(vXL) + FTS(vXM) + FTS(vXR));
//DBW(16, 'F=        ' + FTS(vFL) + FTS(vFM) + FTS(vFR));

            nExitCode :=0;
            exit;
            end;

         // Count the good steps which made good progress
         inc(nGood);
         if (abs(vDX) >= 0.25 * abs(vXStepSize)) then nGood:=0;

//DBW(16, 'Parabol min, vXRec=' + FTS(vXMin + vDX) + FTS(vDX) + ITS5(nGood));
         end;
      end
   else begin // else of : if (abs(vXStepSize) > vTolAbs*0.5) then begin
      // No parabolic fit can be done,
      // take the golden section in the larger of
      // the intervals [vXL,vXMin]; [vXMin,vXR]
      if (vXMin >= vXMiddle) then vXStepSize:= vXL-vXMin
      else                        vXStepSize:= vXR-vXMin;
      vDX := 0.381966*vXStepSize;
//DBW(16, 'Golden, vXRec=' + FTS(vXMin + vDX));
      nGood:=0;
      end;

   // Do not evaluate too near from vXMin.
   if abs(vDX) >= vTolEps then vXRecent:= vXMin + vDX // normal case
   else if (vDX < 0)      then vXRecent:= vXMin - vTolEps
        else                   vXRecent:= vXMin + vTolEps;

   // New function evaluation
   vFRecent := Func(pDataFunc, vXRecent);     inc(nNbFuncEval);

//DBW(16, FTS(vXL) + FTS(vXRecent) + FTS(vXMin) + FTS(vXR) + FTS(vFRecent));

   // Depending on the result of the function evaluation,
   // store intermediate results
   if (vFRecent <= vFMin) then begin
      // A better minimum has been found
      // Shrink the bracket interval
      if (vXRecent >= vXMin) then begin
         vXL:= vXMin;   vFL:=vFMin;
         end
       else begin
         vXR:= vXMin;   vFR:=vFMin;
         end;

//DBW(16, 'Better min found');

      vXMinSecLast := vXMinSec;
      vFMinSecLast := vFMinSec;
      vXMinSec := vXMin;
      vFMinSec := vFMin;
      vXMin := vXRecent;
      vFMin := vFRecent
      end
   else begin // vFRecent > vFMin
      // The last function evaluation didn't give a better minimum,
      // but shrink the bracket interval.
      if (vXRecent < vXMin) then begin
         vXL:= vXRecent;   vFL:=vFRecent;
         end
      else begin
         vXR:= vXRecent;   vFR:=vFRecent;
         end;

//DBW(16, 'No Better min found');

      if (vFRecent <= vFMinSec) or (vXMinSec = vXMin) then begin
         // Last function evaluation is better than second best minimum
         vXMinSecLast := vXMinSec;
         vFMinSecLast := vFMinSec;
         vXMinSec := vXRecent;
         vFMinSec := vFRecent
         end
      else if (vFRecent <= vFMinSecLast) or (vXMinSecLast = vXMin) or
              (vXMinSecLast = vXMinSec) then begin
         // At the starting, or if we got a better minimum than the
         // third last minimum, keept it
         vXMinSecLast := vXRecent;
         vFMinSecLast := vFRecent
         end;
      end;
   until (nNbFuncEval >= nNbFuncEvalMax);

// We didn't find a minimum before reaching the maximum function evaluation.
nExitCode :=1;
vXM  := vXMin;
vFM  := vFMin;
vXSol:= vXMin;
vFSol:= vFMin;
vXFMin:=vXMin;
end;  //  TMinR1SearchR1.SearchMinBrent

end. // unit xMinR1SearchR1;
